@@ -30,7 +30,9 @@ import mock
 
 import networking_calico.plugins.ml2.drivers.calico.test.lib as lib
 
+from networking_calico.common import config as calico_config
 from networking_calico import datamodel_v1
+from networking_calico import datamodel_v2
 from networking_calico import etcdv3
 from networking_calico.monotonic import monotonic_time
 from networking_calico.plugins.ml2.drivers.calico import mech_calico
@@ -64,6 +66,7 @@ class _TestEtcdBase(lib.Lib, unittest.TestCase):
         self.clientv3.delete.side_effect = self.etcd3_delete
         self.clientv3.get.side_effect = self.etcd3_get
         self.clientv3.get_prefix.side_effect = self.etcd3_get_prefix
+        self.clientv3.delete_prefix.side_effect = self.etcd3_delete_prefix
         self.clientv3.status.return_value = {
             'header': {'revision': '10', 'cluster_id': '1234abcd'},
         }
@@ -221,6 +224,14 @@ class _TestEtcdBase(lib.Lib, unittest.TestCase):
         except lib.EtcdKeyNotFound:
             return False
 
+    def etcd3_delete_prefix(self, prefix):
+        _log.info("etcd3 delete prefix: %s", prefix)
+        for key, value in self.etcd_data.items():
+            if key.startswith(prefix):
+                del self.etcd_data[key]
+                _log.info("etcd3 deleted %s", key)
+                self.recent_deletes.add(key)
+
     def etcd3_transaction(self, txn):
         if 'request_put' in txn['success'][0]:
             put_request = txn['success'][0]['request_put']
@@ -238,11 +249,12 @@ class _TestEtcdBase(lib.Lib, unittest.TestCase):
 
         # Slow down reading from etcd status subtree to allow threads to run
         # more often
-        if wait and key == datamodel_v1.FELIX_STATUS_DIR:
+        if wait and key == datamodel_v2.felix_status_dir():
             eventlet.sleep(30)
             self.driver.db.create_or_update_agent = mock.Mock()
 
-        self.etcd_data[datamodel_v1.FELIX_STATUS_DIR + "/vm1/status"] = \
+        self.etcd_data[datamodel_v2.felix_status_dir() +
+                       "/vm1/status"] = \
             json.dumps({
                 "time": "2015-08-14T10:37:54"
             })
@@ -291,15 +303,21 @@ class _TestEtcdBase(lib.Lib, unittest.TestCase):
         return read_result
 
 
-class TestPluginEtcd(_TestEtcdBase):
-    # Tests for the Calico mechanism driver.  This covers the mainline function
-    # and the periodic resync thread.
+class TestPluginEtcdBase(_TestEtcdBase):
+
+    def setUp_region(self):
+        self.region = None
+        self.region_string = "no-region"
+        self.namespace = "openstack"
 
     def setUp(self):
         """Setup before each test case."""
+        self.setUp_region()
+        _log.info("Region %r string %r", self.region, self.region_string)
+
         # Do common plugin test setup.
         lib.m_compat.cfg.CONF.core_plugin = 'ml2'
-        super(TestPluginEtcd, self).setUp()
+        super(TestPluginEtcdBase, self).setUp()
 
         # Mock out the status updating thread.  These tests were originally
         # written before that was added and they do not support the interleaved
@@ -318,74 +336,62 @@ class TestPluginEtcd(_TestEtcdBase):
         lib.m_compat.cfg.CONF.calico.num_port_status_threads = 4
         lib.m_compat.cfg.CONF.calico.etcd_compaction_period_mins = 0
         lib.m_compat.cfg.CONF.calico.project_name_cache_max = 0
+        lib.m_compat.cfg.CONF.calico.openstack_region = self.region
+        calico_config._reset_globals()
+        datamodel_v2._reset_globals()
 
         # This value needs to be a string:
         lib.m_compat.cfg.CONF.keystone_authtoken.auth_url = ""
 
-    sg_default_key_v3 = (
-        '/calico/resources/v3/projectcalico.org/networkpolicies/' +
-        'openstack/ossg.default.SGID-default')
-    sg_default_value_v3 = {
-        'apiVersion': 'projectcalico.org/v3',
-        'kind': 'NetworkPolicy',
-        'metadata': {
-            'namespace': 'openstack',
-            'name': 'ossg.default.SGID-default'
-        },
-        'spec': {
-            'egress': [
-                {'action': 'Allow',
-                 'ipVersion': 4},
-                {'action': 'Allow',
-                 'ipVersion': 6}],
-            'ingress': [
-                {'action': 'Allow',
-                 'ipVersion': 4,
-                 'source': {
-                     'selector':
-                     'has(sg.projectcalico.org/openstack-SGID-default)'}},
-                {'action': 'Allow',
-                 'ipVersion': 6,
-                 'source': {
-                     'selector':
-                     'has(sg.projectcalico.org/openstack-SGID-default)'}}
-            ],
-            'selector': 'has(sg.projectcalico.org/openstack-SGID-default)'}}
-
-    initial_etcd3_writes = {
-        '/calico/resources/v3/projectcalico.org/clusterinformations/default': {
+        self.sg_default_key_v3 = (
+            '/calico/resources/v3/projectcalico.org/networkpolicies/' +
+            self.namespace + '/ossg.default.SGID-default')
+        self.sg_default_value_v3 = {
             'apiVersion': 'projectcalico.org/v3',
-            'kind': 'ClusterInformation',
-            'metadata': {'name': 'default'},
-            'spec': {'clusterGUID': 'uuid-start-no-ports',
-                     'clusterType': 'openstack',
-                     'datastoreReady': True}},
-        '/calico/resources/v3/projectcalico.org/felixconfigurations/default': {
-            'apiVersion': 'projectcalico.org/v3',
-            'kind': 'FelixConfiguration',
-            'metadata': {'name': 'default'},
-            'spec': {'endpointReportingEnabled': True,
-                     'interfacePrefix': 'tap'}},
-        sg_default_key_v3: sg_default_value_v3
-    }
+            'kind': 'NetworkPolicy',
+            'metadata': {
+                'namespace': self.namespace,
+                'name': 'ossg.default.SGID-default'
+            },
+            'spec': {
+                'egress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4},
+                    {'action': 'Allow',
+                     'ipVersion': 6}],
+                'ingress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-SGID-default)'}},
+                    {'action': 'Allow',
+                     'ipVersion': 6,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-SGID-default)'}}
+                ],
+                'selector':
+                'has(sg.projectcalico.org/openstack-SGID-default)'}}
 
-    def test_start_no_ports(self):
-        """Startup with no ports or existing etcd data."""
-        # Allow the etcd transport's resync thread to run. The last thing it
-        # does is write the Felix config, so let it run three reads.
-
-        with lib.FixedUUID('uuid-start-no-ports'):
-            self.give_way()
-            self.simulated_time_advance(31)
-
-        self.assertEtcdWrites(self.initial_etcd3_writes)
-
-    def test_etcd_reset(self):
-        for n in range(1, 20):
-            _log.info("Reset etcd data after %s reads/writes/deletes", n)
-            self.reset_etcd_after = n
-            self.test_start_two_ports()
-            self.etcd_data = {}
+        self.initial_etcd3_writes = {
+            '/calico/resources/v3/projectcalico.org/' +
+            'clusterinformations/default': {
+                'apiVersion': 'projectcalico.org/v3',
+                'kind': 'ClusterInformation',
+                'metadata': {'name': 'default'},
+                'spec': {'clusterGUID': 'uuid-start-no-ports',
+                         'clusterType': 'openstack',
+                         'datastoreReady': True}},
+            '/calico/resources/v3/projectcalico.org/' +
+            'felixconfigurations/default': {
+                'apiVersion': 'projectcalico.org/v3',
+                'kind': 'FelixConfiguration',
+                'metadata': {'name': 'default'},
+                'spec': {'endpointReportingEnabled': True,
+                         'interfacePrefix': 'tap'}},
+            self.sg_default_key_v3: self.sg_default_value_v3
+        }
 
     def make_context(self):
         context = mock.MagicMock()
@@ -404,11 +410,11 @@ class TestPluginEtcd(_TestEtcdBase):
 
         ep_deadbeef_key_v3 = (
             '/calico/resources/v3/projectcalico.org/workloadendpoints/' +
-            'openstack/felix--host--1-openstack-' +
+            self.namespace + '/felix--host--1-openstack-' +
             'instance--1-DEADBEEF--1234--5678')
         ep_facebeef_key_v3 = (
             '/calico/resources/v3/projectcalico.org/workloadendpoints/' +
-            'openstack/felix--host--1-openstack-' +
+            self.namespace + '/felix--host--1-openstack-' +
             'instance--2-FACEBEEF--1234--5678')
         ep_deadbeef_value_v3 = {
             'apiVersion': 'projectcalico.org/v3',
@@ -420,15 +426,16 @@ class TestPluginEtcd(_TestEtcdBase):
                 },
                 'name': ('felix--host--1-openstack-instance' +
                          '--1-DEADBEEF--1234--5678'),
-                'namespace': 'openstack',
+                'namespace': self.namespace,
                 'labels': {
                     'sg.projectcalico.org/openstack-SGID-default':
                     'My_default_SG',
                     'sg-name.projectcalico.org/openstack-My_default_SG':
                     'SGID-default',
-                    'projectcalico.org/namespace': 'openstack',
+                    'projectcalico.org/namespace': self.namespace,
                     'projectcalico.org/openstack-project-id': 'jane3',
                     'projectcalico.org/openstack-project-name': 'pname_jane3',
+                    'projectcalico.org/openstack-project-parent-id': 'gibson',
                     'projectcalico.org/orchestrator': 'openstack'
                 }
             },
@@ -452,15 +459,16 @@ class TestPluginEtcd(_TestEtcdBase):
                 },
                 'name': ('felix--host--1-openstack-instance' +
                          '--2-FACEBEEF--1234--5678'),
-                'namespace': 'openstack',
+                'namespace': self.namespace,
                 'labels': {
                     'sg.projectcalico.org/openstack-SGID-default':
                     'My_default_SG',
                     'sg-name.projectcalico.org/openstack-My_default_SG':
                     'SGID-default',
-                    'projectcalico.org/namespace': 'openstack',
+                    'projectcalico.org/namespace': self.namespace,
                     'projectcalico.org/openstack-project-id': 'jane3',
                     'projectcalico.org/openstack-project-name': 'pname_jane3',
+                    'projectcalico.org/openstack-project-parent-id': 'gibson',
                     'projectcalico.org/orchestrator': 'openstack'
                 }
             },
@@ -564,7 +572,7 @@ class TestPluginEtcd(_TestEtcdBase):
 
         ep_hello_key_v3 = (
             '/calico/resources/v3/projectcalico.org/workloadendpoints/' +
-            'openstack/felix--host--2-openstack-' +
+            self.namespace + '/felix--host--2-openstack-' +
             'instance--3-HELLO--1234--5678')
         ep_hello_value_v3 = {
             'apiVersion': 'projectcalico.org/v3',
@@ -576,15 +584,16 @@ class TestPluginEtcd(_TestEtcdBase):
                 },
                 'name': ('felix--host--2-openstack-instance' +
                          '--3-HELLO--1234--5678'),
-                'namespace': 'openstack',
+                'namespace': self.namespace,
                 'labels': {
                     'sg.projectcalico.org/openstack-SGID-default':
                     'My_default_SG',
                     'sg-name.projectcalico.org/openstack-My_default_SG':
                     'SGID-default',
-                    'projectcalico.org/namespace': 'openstack',
+                    'projectcalico.org/namespace': self.namespace,
                     'projectcalico.org/openstack-project-id': 'jane3',
                     'projectcalico.org/openstack-project-name': 'pname_jane3',
+                    'projectcalico.org/openstack-project-parent-id': 'gibson',
                     'projectcalico.org/orchestrator': 'openstack'
                 }
             },
@@ -705,12 +714,12 @@ class TestPluginEtcd(_TestEtcdBase):
 
         sg_1_key_v3 = (
             '/calico/resources/v3/projectcalico.org/networkpolicies/' +
-            'openstack/ossg.default.SG-1')
+            self.namespace + '/ossg.default.SG-1')
         sg_1_value_v3 = {
             'apiVersion': 'projectcalico.org/v3',
             'kind': 'NetworkPolicy',
             'metadata': {
-                'namespace': 'openstack',
+                'namespace': self.namespace,
                 'name': 'ossg.default.SG-1'
             },
             'spec': {
@@ -883,6 +892,29 @@ class TestPluginEtcd(_TestEtcdBase):
             'port_range_max': 5060
         }
 
+
+class TestPluginEtcd(TestPluginEtcdBase):
+    # Tests for the Calico mechanism driver.  This covers the mainline function
+    # and the periodic resync thread.
+
+    def test_start_no_ports(self):
+        """Startup with no ports or existing etcd data."""
+        # Allow the etcd transport's resync thread to run. The last thing it
+        # does is write the Felix config, so let it run three reads.
+
+        with lib.FixedUUID('uuid-start-no-ports'):
+            self.give_way()
+            self.simulated_time_advance(31)
+
+        self.assertEtcdWrites(self.initial_etcd3_writes)
+
+    def test_etcd_reset(self):
+        for n in range(1, 20):
+            _log.info("Reset etcd data after %s reads/writes/deletes", n)
+            self.reset_etcd_after = n
+            self.test_start_two_ports()
+            self.etcd_data = {}
+
     def test_noop_entry_points(self):
         """test_noop_entry_points
 
@@ -898,6 +930,17 @@ class TestPluginEtcd(_TestEtcdBase):
     def test_subnet_hooks(self):
         """Test subnet creation, update and deletion hooks."""
 
+        # Simulate some legacy (pre-multi-region) subnet data.
+        self.etcd_data = {
+            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24': json.dumps({
+                'network_id': 'net-id-1',
+                'cidr': '10.65.0.0/24',
+                'gateway_ip': '10.65.0.1',
+                'host_routes': [{'destination': '11.11.0.0/16',
+                                 'nexthop': '10.65.0.1'}]
+            }),
+        }
+
         # Allow the etcd transport's resync thread to run.  Expect the usual
         # writes.
         with lib.FixedUUID('uuid-subnet-hooks'):
@@ -910,6 +953,11 @@ class TestPluginEtcd(_TestEtcdBase):
             'default']['spec']['clusterGUID'] = 'uuid-subnet-hooks'
         self.assertEtcdWrites(expected_writes)
 
+        # Check that the legacy data has been cleaned up.
+        self.assertEtcdDeletes(set([
+            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24',
+        ]))
+
         # Define two subnets.
         subnet1 = {'network_id': 'net-id-1',
                    'enable_dhcp': True,
@@ -933,7 +981,7 @@ class TestPluginEtcd(_TestEtcdBase):
         context.current = subnet1
         self.driver.create_subnet_postcommit(context)
         self.assertEtcdWrites({
-            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24': {
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.65.0--24': {
                 'network_id': 'net-id-1',
                 'cidr': '10.65.0.0/24',
                 'gateway_ip': '10.65.0.1',
@@ -960,7 +1008,7 @@ class TestPluginEtcd(_TestEtcdBase):
         context.current = subnet1
         self.driver.update_subnet_postcommit(context)
         self.assertEtcdDeletes(set([
-            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24'
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.65.0--24'
         ]))
 
         # Update subnet2 to be DHCP-enabled.
@@ -968,7 +1016,7 @@ class TestPluginEtcd(_TestEtcdBase):
         context.current = subnet2
         self.driver.update_subnet_postcommit(context)
         self.assertEtcdWrites({
-            '/calico/dhcp/v1/subnet/subnet-id-10.28.0--24': {
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.28.0--24': {
                 'network_id': 'net-id-2',
                 'cidr': '10.28.0.0/24',
                 'gateway_ip': '10.28.0.1',
@@ -987,7 +1035,7 @@ class TestPluginEtcd(_TestEtcdBase):
             '/calico/resources/v3/projectcalico.org/clusterinformations/' +
             'default']['spec']['clusterGUID'] = 'uuid-subnet-hooks-2'
         expected_writes.update({
-            '/calico/dhcp/v1/subnet/subnet-id-10.28.0--24': {
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.28.0--24': {
                 'network_id': 'net-id-2',
                 'cidr': '10.28.0.0/24',
                 'gateway_ip': '10.28.0.1',
@@ -1004,7 +1052,7 @@ class TestPluginEtcd(_TestEtcdBase):
         self.give_way()
         self.simulated_time_advance(mech_calico.RESYNC_INTERVAL_SECS)
         self.assertEtcdWrites({
-            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24': {
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.65.0--24': {
                 'network_id': 'net-id-1',
                 'cidr': '10.65.0.0/24',
                 'gateway_ip': '10.65.0.1',
@@ -1012,7 +1060,7 @@ class TestPluginEtcd(_TestEtcdBase):
                                  'nexthop': '10.65.0.1'}]}
         })
         self.assertEtcdDeletes(set([
-            '/calico/dhcp/v1/subnet/subnet-id-10.28.0--24'
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.28.0--24'
         ]))
 
         # Do a resync where we simulate having missed a dynamic update that
@@ -1021,7 +1069,7 @@ class TestPluginEtcd(_TestEtcdBase):
         self.give_way()
         self.simulated_time_advance(mech_calico.RESYNC_INTERVAL_SECS)
         self.assertEtcdWrites({
-            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24': {
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.65.0--24': {
                 'network_id': 'net-id-1',
                 'cidr': '10.65.0.0/24',
                 'gateway_ip': '10.65.0.2',
@@ -1040,7 +1088,7 @@ class TestPluginEtcd(_TestEtcdBase):
         context.current = subnet1
         self.driver.delete_subnet_postcommit(context)
         self.assertEtcdDeletes(set([
-            '/calico/dhcp/v1/subnet/subnet-id-10.65.0--24'
+            '/calico/dhcp/v2/no-region/subnet/subnet-id-10.65.0--24'
         ]))
 
     def test_check_segment_for_agent(self):
@@ -1172,13 +1220,14 @@ class TestPluginEtcd(_TestEtcdBase):
         self.assertEtcdDeletes(set())
 
     def test_policy_coexistence(self):
-        """Coexistence with other policy data in the 'openstack' namespace.
+        """Coexistence with other data in the 'openstack' namespace.
 
         Check that we _do_ clean up old policy data that has our prefix, but
         _don't_ touch policies without our prefix.
         """
 
-        # Start up with two existing policies in the 'openstack' namespace.
+        # Start up with two existing policies in the 'openstack'
+        # namespace.
         self.etcd_data = {
             '/calico/resources/v3/projectcalico.org/networkpolicies/' +
             'openstack/customer-policy-1': json.dumps({
@@ -1225,35 +1274,101 @@ class TestPluginEtcd(_TestEtcdBase):
     def test_old_openstack_data(self):
         """Startup with existing but old OpenStack profile data."""
 
-        # Check that we clean it up.
+        # Check that we clean up policy data that we created, but not policy
+        # data that the user created.
+        our_policy_string = json.dumps({
+            'apiVersion': 'projectcalico.org/v3',
+            'kind': 'NetworkPolicy',
+            'metadata': {
+                'namespace': 'openstack',
+                'name': 'ossg.default.OLD'
+            },
+            'spec': {
+                'egress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4},
+                    {'action': 'Allow',
+                     'ipVersion': 6}],
+                'ingress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-OLD)'}},
+                    {'action': 'Allow',
+                     'ipVersion': 6,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-OLD)'}}],
+                'selector': 'has(sg.projectcalico.org/openstack-OLD)'}}
+        )
+        user_policy_string = our_policy_string.replace("ossg", "user")
         self.etcd_data = {
             '/calico/resources/v3/projectcalico.org/networkpolicies/' +
-            'openstack/ossg.default.OLD': json.dumps({
-                'apiVersion': 'projectcalico.org/v3',
-                'kind': 'NetworkPolicy',
-                'metadata': {
-                    'namespace': 'openstack',
-                    'name': 'ossg.default.OLD'
-                },
-                'spec': {
-                    'egress': [
-                        {'action': 'Allow',
-                         'ipVersion': 4},
-                        {'action': 'Allow',
-                         'ipVersion': 6}],
-                    'ingress': [
-                        {'action': 'Allow',
-                         'ipVersion': 4,
-                         'source': {
-                             'selector':
-                             'has(sg.projectcalico.org/openstack-OLD)'}},
-                        {'action': 'Allow',
-                         'ipVersion': 6,
-                         'source': {
-                             'selector':
-                             'has(sg.projectcalico.org/openstack-OLD)'}}],
-                    'selector': 'has(sg.projectcalico.org/openstack-OLD)'}}
-            )}
+            'openstack/ossg.default.OLD': our_policy_string,
+            '/calico/resources/v3/projectcalico.org/networkpolicies/' +
+            'openstack/user.default.OLD': user_policy_string,
+        }
+        with lib.FixedUUID('uuid-old-data'):
+            self.give_way()
+            self.simulated_time_advance(31)
+
+        expected_writes = copy.deepcopy(self.initial_etcd3_writes)
+        expected_writes[
+            '/calico/resources/v3/projectcalico.org/clusterinformations/' +
+            'default']['spec']['clusterGUID'] = 'uuid-old-data'
+        self.assertEtcdWrites(expected_writes)
+        self.assertEtcdDeletes(set([
+            '/calico/resources/v3/projectcalico.org/networkpolicies/' +
+            'openstack/ossg.default.OLD',
+        ]))
+
+
+class TestPluginEtcdRegion(TestPluginEtcdBase):
+
+    def setUp_region(self):
+        self.region = "europe"
+        self.region_string = "region-europe"
+        self.namespace = "openstack-region-europe"
+
+    def test_legacy_openstack_data(self):
+        """Startup with existing data in legacy "openstack" namespace."""
+
+        # Check that we clean up policy data that we created, but not policy
+        # data that the user created.
+        our_policy_string = json.dumps({
+            'apiVersion': 'projectcalico.org/v3',
+            'kind': 'NetworkPolicy',
+            'metadata': {
+                'namespace': 'openstack',
+                'name': 'ossg.default.OLD'
+            },
+            'spec': {
+                'egress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4},
+                    {'action': 'Allow',
+                     'ipVersion': 6}],
+                'ingress': [
+                    {'action': 'Allow',
+                     'ipVersion': 4,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-OLD)'}},
+                    {'action': 'Allow',
+                     'ipVersion': 6,
+                     'source': {
+                         'selector':
+                         'has(sg.projectcalico.org/openstack-OLD)'}}],
+                'selector': 'has(sg.projectcalico.org/openstack-OLD)'}}
+        )
+        user_policy_string = our_policy_string.replace("ossg", "user")
+        self.etcd_data = {
+            '/calico/resources/v3/projectcalico.org/networkpolicies/' +
+            'openstack/ossg.default.OLD': our_policy_string,
+            '/calico/resources/v3/projectcalico.org/networkpolicies/' +
+            'openstack/user.default.OLD': user_policy_string,
+        }
         with lib.FixedUUID('uuid-old-data'):
             self.give_way()
             self.simulated_time_advance(31)
@@ -1495,18 +1610,48 @@ class TestDriverStatusReporting(lib.Lib, unittest.TestCase):
         self.assertEqual([mock.call(("host", "port"))], m_queue.put.mock_calls)
 
 
-class TestStatusWatcher(_TestEtcdBase):
+class TestStatusWatcherBase(_TestEtcdBase):
+
+    def setUp_region(self):
+        self.region = None
+        self.region_string = "no-region"
+
     def setUp(self):
+        self.setUp_region()
+        _log.info("Region %r string %r", self.region, self.region_string)
+
         # Mock out config.
         lib.m_compat.cfg.CONF.calico.etcd_host = "localhost"
         lib.m_compat.cfg.CONF.calico.etcd_port = 4001
         lib.m_compat.cfg.CONF.calico.etcd_key_file = None
         lib.m_compat.cfg.CONF.calico.etcd_cert_file = None
         lib.m_compat.cfg.CONF.calico.etcd_ca_cert_file = None
-        super(TestStatusWatcher, self).setUp()
+        lib.m_compat.cfg.CONF.calico.openstack_region = self.region
+        calico_config._reset_globals()
+        datamodel_v2._reset_globals()
+
+        super(TestStatusWatcherBase, self).setUp()
         self.driver = mock.Mock(spec=mech_calico.CalicoMechanismDriver)
         self.watcher = status.StatusWatcher(self.driver)
 
+    def _add_test_endpoint(self):
+        # Add a workload to be deleted
+        m_port_status_node = mock.Mock()
+        m_port_status_node.key = (
+            ("/calico/felix/v2/%s/host/hostname/workload/" +
+             "openstack/wlid/endpoint/ep1") % self.region_string
+        )
+        m_port_status_node.value = '{"status": "up"}'
+        self.watcher._on_ep_set(m_port_status_node, "hostname", "wlid", "ep1")
+        ep_id = datamodel_v1.WloadEndpointId("hostname", "openstack",
+                                             "wlid", "ep1")
+        self.assertEqual({"hostname": set([ep_id])},
+                         self.watcher._endpoints_by_host)
+        return m_port_status_node
+
+
+class TestStatusWatcher(TestStatusWatcherBase):
+
     def test_tls(self):
         lib.m_compat.cfg.CONF.calico.etcd_cert_file = "cert-file"
         lib.m_compat.cfg.CONF.calico.etcd_ca_cert_file = "ca-cert-file"
@@ -1516,13 +1661,15 @@ class TestStatusWatcher(_TestEtcdBase):
     def test_snapshot(self):
         # Populate initial status tree data, for initial snapshot testing.
 
-        felix_status_key = '/calico/felix/v1/host/hostname/status'
+        felix_status_key = '/calico/felix/v2/no-region/host/hostname/status'
         felix_last_reported_status_key = \
-            '/calico/felix/v1/host/hostname/last_reported_status'
-        ep_on_that_host_key = ('/calico/felix/v1/host/hostname/workload/' +
-                               'openstack/wlid/endpoint/ep1')
-        ep_on_unknown_host_key = ('/calico/felix/v1/host/unknown/workload/' +
-                                  'openstack/wlid/endpoint/ep2')
+            '/calico/felix/v2/no-region/host/hostname/last_reported_status'
+        ep_on_that_host_key = (
+            '/calico/felix/v2/no-region/host/hostname/workload/' +
+            'openstack/wlid/endpoint/ep1')
+        ep_on_unknown_host_key = (
+            '/calico/felix/v2/no-region/host/unknown/workload/' +
+            'openstack/wlid/endpoint/ep2')
 
         self.etcd_data = {
             # An agent status key to ignore.
@@ -1605,8 +1752,8 @@ class TestStatusWatcher(_TestEtcdBase):
         # back to high after the snapshot.
         watch_events = [{
             "kv": {
-                "key": "/calico/felix/v1/host/hostname/workload/openstack/"
-                       "wlid/endpoint/ep1",
+                "key": "/calico/felix/v2/no-region/host/hostname/workload/"
+                       "openstack/wlid/endpoint/ep1",
                 "value": '{"status": "up"}',
             },
             "type": "SET",
@@ -1637,7 +1784,7 @@ class TestStatusWatcher(_TestEtcdBase):
 
     def test_endpoint_status_add_bad_json(self):
         m_port_status_node = mock.Mock()
-        m_port_status_node.key = "/calico/felix/v1/host/hostname/workload/" \
+        m_port_status_node.key = "/calico/felix/v2/no-region/host/hostname/workload/" \
                                  "openstack/wlid/endpoint/ep1"
         m_port_status_node.value = '{"status": "up"'
         self.watcher._on_ep_set(m_port_status_node, "hostname", "wlid", "ep1")
@@ -1651,7 +1798,7 @@ class TestStatusWatcher(_TestEtcdBase):
 
     def test_endpoint_status_add_bad_id(self):
         m_port_status_node = mock.Mock()
-        m_port_status_node.key = "/calico/felix/v1/host/hostname/workload/" \
+        m_port_status_node.key = "/calico/felix/v2/no-region/host/hostname/workload/" \
                                  "openstack/wlid/endpoint"
         self.watcher._on_ep_set(m_port_status_node, "hostname", "wlid", "ep1")
         self.assertEqual(
@@ -1659,23 +1806,10 @@ class TestStatusWatcher(_TestEtcdBase):
             self.driver.on_port_status_changed.mock_calls)
         self.assertEqual({}, self.watcher._endpoints_by_host)
 
-    def _add_test_endpoint(self):
-        # Add a workload to be deleted
-        m_port_status_node = mock.Mock()
-        m_port_status_node.key = "/calico/felix/v1/host/hostname/workload/" \
-                                 "openstack/wlid/endpoint/ep1"
-        m_port_status_node.value = '{"status": "up"}'
-        self.watcher._on_ep_set(m_port_status_node, "hostname", "wlid", "ep1")
-        ep_id = datamodel_v1.WloadEndpointId("hostname", "openstack",
-                                             "wlid", "ep1")
-        self.assertEqual({"hostname": set([ep_id])},
-                         self.watcher._endpoints_by_host)
-        return m_port_status_node
-
     def test_status_bad_json(self):
         for value in ["{", 10, "foo"]:
             m_response = mock.Mock()
-            m_response.key = "/calico/felix/v1/host/hostname/status"
+            m_response.key = "/calico/felix/v2/no-region/host/hostname/status"
             m_response.value = value
             self.watcher._on_status_set(m_response, "foo")
         self.assertFalse(self.driver.on_felix_alive.called)
@@ -1683,14 +1817,14 @@ class TestStatusWatcher(_TestEtcdBase):
     def test_felix_status_expiry(self):
         # Put an endpoint in the cache to find later...
         m_response = mock.Mock()
-        m_response.key = "/calico/felix/v1/host/hostname/workload/" \
+        m_response.key = "/calico/felix/v2/no-region/host/hostname/workload/" \
                          "openstack/wlid/endpoint/epid"
         m_response.value = '{"status": "up"}'
         self.watcher._on_ep_set(m_response, "hostname", "wlid", "epid")
 
         # Then note that felix is down.
         m_response = mock.Mock()
-        m_response.key = "/calico/felix/v1/host/hostname/status"
+        m_response.key = "/calico/felix/v2/no-region/host/hostname/status"
         self.watcher._on_status_del(m_response, "hostname")
 
         # Check that nothing happens to the port.  (Previously, we used to mark
@@ -1704,6 +1838,85 @@ class TestStatusWatcher(_TestEtcdBase):
             self.driver.on_port_status_changed.mock_calls)
 
 
+class TestMultiRegionStatusWatcher(TestStatusWatcherBase):
+
+    def setUp_region(self):
+        self.region = "europe"
+        self.region_string = "region-europe"
+
+    def test_endpoint_status_add_delete(self):
+        m_port_status_node = self._add_test_endpoint()
+        m_port_status_node.action = "delete"
+        self.watcher._on_ep_delete(m_port_status_node,
+                                   "hostname", "wlid", "ep1")
+
+        self.assertEqual(
+            [
+                mock.call("hostname", "ep1", {"status": "up"},
+                          priority="high"),
+                mock.call("hostname", "ep1", None, priority="high"),
+            ],
+            self.driver.on_port_status_changed.mock_calls)
+        self.assertEqual({}, self.watcher._endpoints_by_host)
+
+    def test_handle_port_this_region(self):
+        # Simulate status update for a workload in this region.
+        m_port_status_node = mock.Mock()
+        m_port_status_node.key = (
+            "/calico/felix/v2/" + self.region_string +
+            "/host/hostname/workload/openstack/wlid/endpoint/ep1"
+        )
+        m_port_status_node.value = '{"status": "up"}'
+        m_port_status_node.action = "set"
+        self.watcher.dispatcher.handle_event(m_port_status_node)
+        self.assertEqual(
+            [
+                mock.call("hostname", "ep1", {"status": "up"},
+                          priority="high"),
+            ],
+            self.driver.on_port_status_changed.mock_calls)
+
+    def test_ignore_port_other_region(self):
+        # Simulate status update for a workload in another region.
+        m_port_status_node = mock.Mock()
+        m_port_status_node.key = (
+            "/calico/felix/v2/region-other/host/hostname/workload/" +
+            "openstack/wlid/endpoint/ep1"
+        )
+        m_port_status_node.value = '{"status": "up"}'
+        m_port_status_node.action = "set"
+        self.watcher.dispatcher.handle_event(m_port_status_node)
+        self.assertEqual(
+            [],
+            self.driver.on_port_status_changed.mock_calls)
+
+    def test_handle_felix_this_region(self):
+        self.driver.on_felix_alive.reset_mock()
+        m_response = mock.Mock()
+        m_response.action = "set"
+        m_response.key = ("/calico/felix/v2/" +
+                          self.region_string +
+                          "/host/hostname/status")
+        m_response.value = json.dumps({
+            "uptime": 10,
+            "first_update": True,
+        })
+        self.watcher.dispatcher.handle_event(m_response)
+        self.assertTrue(self.driver.on_felix_alive.called)
+
+    def test_ignore_felix_other_region(self):
+        self.driver.on_felix_alive.reset_mock()
+        m_response = mock.Mock()
+        m_response.action = "set"
+        m_response.key = "/calico/felix/v2/region-other/host/hostname/status"
+        m_response.value = json.dumps({
+            "uptime": 10,
+            "first_update": True,
+        })
+        self.watcher.dispatcher.handle_event(m_response)
+        self.assertFalse(self.driver.on_felix_alive.called)
+
+
 def _neutron_rule_from_dict(overrides):
     rule = {
         "ethertype": "IPv4",
